#!/usr/bin/env python
# encoding: utf-8
"""
__init__.py - __init__ file for the inpainter module
~ Daniel Cortild, 21 March 2023
"""

# from .Image import Image
# from .Mask import Mask
# from .Algorithm import Algorithm
from .experiments import plotExperiments, plotItsTime, plotExperimentRegularisation
from .getInpainted import getInpainted